export interface IteratorResult<D> {
    node: Node<D>;
    remainingPath: string;
}
export declare class Node<D> {
    /** Path segment this Node represents. */
    segment: string;
    /** Children of this Node. */
    children: Node<D>[];
    /** Function used to initialize the data object. */
    dataCreator: () => D;
    /** Data stored in this Node. */
    data: D;
    /**
     * A radix tree implementation.
     *
     * @param dataCreator Function used to initialize the data variable of new Nodes
     * @param segment Part of the path for this Node
     */
    constructor(dataCreator: () => D, segment?: string);
    /**
     * Returns `true` if this node has no children.
     */
    isLeaf(): boolean;
    /**
     * Generator function yielding for every Node encountered.
     *
     * Similar to `findAll()` but allows stopping at any point (by not invoking the generator) and changing the path on the fly (to handle params).
     *
     * @param path Path of the node to find
     */
    nodeIterator(path: string): Generator<IteratorResult<D>, void, undefined | string>;
    /**
     * Traverses the tree to find all Nodes leading to the given path.
     *
     * If `createIfNone` is set and no Node exists at the exact given path, a new Node will be inserted there, even if it will lead to a Node with just one child. This is used to allow having data in the middle of a path.
     *
     * @param path Path of node to find
     * @param createIfNone If `true`, a Node will be created for the given path if one isn't found
     */
    findAll(path: string, createIfNone: true): Node<D>[];
    findAll(path: string, createIfNone?: boolean): Node<D>[] | null;
    /**
     * Returns the last Node at the exact given path.
     *
     * If `createIfNone` is `true`, inserts a new Node if one doesn't already exist there, then returns it.
     *
     * @param path Path of node to find
     * @param createIfNone If `true`, allows creation of new Nodes
     */
    find(path: string, createIfNone: true): Node<D>;
    find(path: string, createIfNone?: boolean): Node<D> | null;
    /**
     * Returns the tree as a string. Useful for debugging.'
     *
     * @param indent Level of indentation for the returned string. Defaults to 0.
     */
    stringify(indent?: number): string;
    [Symbol.iterator](): Generator<Node<D>, void, unknown>;
    /**
     * Returns the amount of characters the beginnings of both strings share.
     *
     * @param a First string
     * @param b Second string
     */
    static commonLength(a: string, b: string): number;
}
