"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Node = void 0;
class Node {
    /**
     * A radix tree implementation.
     *
     * @param dataCreator Function used to initialize the data variable of new Nodes
     * @param segment Part of the path for this Node
     */
    constructor(dataCreator, segment = '<root>') {
        /** Children of this Node. */
        this.children = [];
        this.segment = segment;
        this.dataCreator = dataCreator;
        this.data = dataCreator();
    }
    /**
     * Returns `true` if this node has no children.
     */
    isLeaf() {
        return this.children.length === 0;
    }
    /**
     * Generator function yielding for every Node encountered.
     *
     * Similar to `findAll()` but allows stopping at any point (by not invoking the generator) and changing the path on the fly (to handle params).
     *
     * @param path Path of the node to find
     */
    *nodeIterator(path) {
        let remainingPath = (yield {
            node: this,
            remainingPath: path,
        }) || path;
        // If path is empty, stop there
        if (remainingPath.length === 0)
            return;
        let node = this;
        while (node !== null && !node.isLeaf() && remainingPath.length > 0) {
            let found = false;
            for (let child of node.children) {
                if (remainingPath.startsWith(child.segment)) {
                    node = child;
                    remainingPath = remainingPath.substr(child.segment.length);
                    const newPath = yield {
                        node,
                        remainingPath,
                    };
                    // If a new path is passed in, replace the remaining path with it.
                    if (typeof newPath === 'string')
                        remainingPath = newPath;
                    // If no more path remains, end the generator
                    if (remainingPath.length === 0)
                        return;
                    // We found our child, iterate over its children
                    found = true;
                    break;
                }
            }
            // If we didn't find any children, we reached the end of the chain and should stop
            if (!found)
                break;
        }
    }
    findAll(path, createIfNone = false) {
        let node = this;
        const output = [node];
        // If path is empty, return the root node
        if (path.length === 0)
            return output;
        let remainingPath = path;
        while (node !== null && !node.isLeaf() && remainingPath.length > 0) {
            let found = false;
            for (let child of node.children) {
                if (remainingPath.startsWith(child.segment)) {
                    node = child;
                    remainingPath = remainingPath.substr(child.segment.length);
                    output.push(node);
                    if (remainingPath.length === 0)
                        return output;
                    found = true;
                    break;
                }
            }
            if (!found)
                break;
        }
        if (!createIfNone)
            return null;
        let bestCommonLength = 0;
        let bestNode = null;
        for (let child of node.children) {
            const commonLength = Node.commonLength(remainingPath, child.segment);
            if (commonLength > 0) {
                bestCommonLength = commonLength;
                bestNode = child;
                break;
            }
        }
        if (!bestNode) {
            // There are no nodes with a matching prefix. Create a new node and return it.
            const newNode = new Node(this.dataCreator, remainingPath);
            node.children.push(newNode);
            output.push(newNode);
            return output;
        }
        const commonSegment = bestNode.segment.substr(0, bestCommonLength);
        // Create a new node for the new branches
        const branchNode = new Node(this.dataCreator, commonSegment);
        output.push(branchNode);
        // Modify the children of the last found node
        node.children.splice(this.children.indexOf(bestNode), 1, branchNode);
        // Adjust the old node
        bestNode.segment = bestNode.segment.substr(bestCommonLength);
        branchNode.children.push(bestNode);
        // The newly created branch node is the last part of the path
        if (remainingPath.length <= commonSegment.length)
            return output;
        // There's still some path remaining. Create a node for the remaining part and return it.
        const newNode = new Node(this.dataCreator, remainingPath.substr(bestCommonLength));
        branchNode.children.push(newNode);
        output.push(newNode);
        return output;
    }
    find(path, createIfNone = false) {
        const nodes = this.findAll(path, createIfNone);
        if (!nodes)
            return null;
        return nodes[nodes.length - 1];
    }
    /**
     * Returns the tree as a string. Useful for debugging.'
     *
     * @param indent Level of indentation for the returned string. Defaults to 0.
     */
    stringify(indent = 0) {
        var _a;
        const output = [];
        const indendStr = new Array(indent).fill('  ').join('');
        output.push(`${indendStr}${this.segment}`);
        if (typeof ((_a = this.data) === null || _a === void 0 ? void 0 : _a.toString) === 'function')
            output.push(`: ${this.data.toString()}`);
        for (const node of this.children) {
            output.push(`\n${node.stringify(indent + 1)}`);
        }
        return output.join('');
    }
    *[Symbol.iterator]() {
        yield this;
        for (const node of this.children) {
            const iterator = node[Symbol.iterator]();
            let result = iterator.next();
            while (!result.done) {
                yield result.value;
                result = iterator.next();
            }
        }
    }
    /**
     * Returns the amount of characters the beginnings of both strings share.
     *
     * @param a First string
     * @param b Second string
     */
    static commonLength(a, b) {
        const maxChars = Math.min(a.length, b.length);
        for (let i = 0; i < maxChars; i++) {
            if (a[i] !== b[i])
                return i;
        }
        return maxChars;
    }
}
exports.Node = Node;
//# sourceMappingURL=Node.js.map