import { DefaultContext, DefaultState, Middleware, Next, ParameterizedContext } from 'koa';
import compose from 'koa-compose';
import { Node } from './Node';
import { StagedArray } from './StagedArray';
export declare const TERMINATOR_MIDDLEWARE_KEY = "koa-butterfly/terminator-middleware";
/**
 * Contains keys used to internally store special types of middleware.
 *
 * Names are prefixed with a null byte because according to the HTTP RFC, these method names are otherwise technically valid, despite Node.js seemingly rejecting them. See https://tools.ietf.org/html/rfc7231#section-4
 */
export declare enum SpecialMethod {
    /**
     * Middleware for this special method will get executed for matching path segments, as soon as they're encountered.
     *
     * Terminators will be executed at the end of the request, but only if the request path matches a node with terminators.
     */
    MIDDLEWARE = "\0middleware",
    /**
     * Middleware bound to this special method is called together with the middleware for the specific HTTP request methods.
     *
     * Terminators bound to this method are called for any HTTP request method, but only after the more specific options are exhausted.
     */
    ALL = "\0all"
}
export declare class MethodData<D> {
    middleware: StagedArray<D>;
    terminators: StagedArray<D>;
}
export interface ParameterRoute<NodeT> {
    name: string;
    regex: RegExp | null;
    matchAll: boolean;
    rootNode: NodeT;
}
export declare class RouterNodeData<StateT, ContextT> {
    /** Stores middleware used for different HTTP methods, as well as some special methods, which are used for mounting middleware on specific paths, and middleware that should run for all methods. */
    methodData: Map<string, MethodData<Middleware<StateT, ContextT>>>;
    lateParams: StagedArray<ParameterRoute<Node<RouterNodeData<StateT, ContextT>>>>;
    getMethodData(method: string, createIfMissing: true): MethodData<Middleware<StateT, ContextT>>;
    getMethodData(method: string, createIfMissing?: boolean): MethodData<Middleware<StateT, ContextT>> | undefined;
    toString(): string;
}
export interface RouterOptions {
    /** If true, requests ending with `/` will match routes not ending in `/`. Routes ending with `/` will still require the request path to end in a slash. */
    strictSlashes?: boolean;
}
export declare type RouterContext<StateT, ContextT> = {
    params: Record<string, string>;
    remainingUrl: string;
    [TERMINATOR_MIDDLEWARE_KEY]: StagedArray<Middleware<StateT, ContextT>>[];
} & ContextT;
export interface Middlewareable {
    middleware(): Middleware<any, any>;
}
export declare type MiddlewareLike<StateT = DefaultState, ContextT = DefaultContext> = Middleware<StateT, ContextT> | Middlewareable | null | undefined | false;
/**
 * Used to set values for all middleware executed before the passed in `next` is called, then reset them back to the old value.
 *
 * In practice, this can be used to set parameter values while in the Router, while keeping them at their old values while giving up control to the `next` passed to the Router's middleware callback.
 *
 * This is similar to what `koa-mount` does with `ctx.path` and `ctx.mountPath`.
 *
 * @param oldValue Value passed to the setter to reset the value back to the old one
 * @param newValue Value passed to the setter to set the new value
 * @param setter Function setting the wanted values
 * @param next The middleware callback to call after the `wrappedNext` passed through `callback` is called
 * @param callback Function to call after the values have been set
 */
export declare function middlewareValueWrapper<T>(oldValue: T, newValue: T, setter: (value: T) => any, next: Next, callback: (wrappedNext: () => any) => Promise<void>): Promise<void>;
/**
 * A router implementation for Koa using a radix tree.
 */
export declare class Router<StateT = DefaultState, ContextT = DefaultContext, RouterContextT extends RouterContext<StateT, ContextT> = RouterContext<StateT, ContextT>> implements Middlewareable {
    rootNode: Node<RouterNodeData<StateT, RouterContextT>>;
    private strictSlashes;
    constructor(opts?: RouterOptions);
    getNode(path: string, createIfNone: true): this['rootNode'];
    getNode(path: string, createIfNone: false): this['rootNode'] | null;
    addMiddleware(method: string, path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    addTerminator(method: string, path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    private addMiddlewareData;
    middleware(): Middleware<StateT, RouterContextT>;
    middlewareHandler(ctx: ParameterizedContext<StateT, RouterContextT>, next: Next): Promise<void>;
    handleNodePath(node: this['rootNode'], path: string, ctx: ParameterizedContext<StateT, RouterContextT>, next: Next): Promise<void>;
    middlewareGenerator(node: this['rootNode'], ctx: ParameterizedContext<StateT, RouterContextT>, next: Next): AsyncGenerator<compose.Middleware<ParameterizedContext<StateT, RouterContextT>>[], any, unknown>;
    use(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    use(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    register(method: string, path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    register(method: string, path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    register(method: string, path: string, stage: number | MiddlewareLike<StateT, RouterContextT>, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    all(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    all(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    connect(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    connect(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    delete(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    delete(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    del(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    del(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    get(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    get(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    head(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    head(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    options(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    options(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    patch(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    patch(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    post(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    post(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    put(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    put(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    trace(path: string, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
    trace(path: string, stage: number, ...middleware: MiddlewareLike<StateT, RouterContextT>[]): this;
}
